/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2004  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *  Copyright (C) 1999-2004  Haavard Kvaalen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * bemused mpd hack
 *
 * v0.1 - initial version by michu@neophob.com 
 * v0.2 - use libmpdclient.h instead of libmpd
 * v0.3 - fix various connection errors
 * v0.4 - remove excessive debug logs, add libiconv function
 *
 */
 
#if defined(HAVE_CONFIG_H)
#include "config.h"
#endif

#include <glib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <iconv.h>

#include "libmpdclient.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "mpdctrl.h"

/* timeout is 15 minutes */
#define TIMEOUTINS 60*15

#define ERROR_DISPLAY 5

static int l_totalTimeSec;
static int l_elapsedTimeSec;
static int l_bitRate;
static int l_repeatEnabled;
static int l_randomEnabled;
static int l_state;
static int l_volume;
static int l_numberOfSongs;
static unsigned long l_uptime;
static unsigned long l_playTime;
static unsigned long l_dbPlayTime;
static int l_playlistLength;
/* pos in playlist */
static int l_currentSongPos;
static unsigned int l_sampleRate;
static int l_channels;

static mpd_Song *currentSong;

/* connection information */
static char host[512];
static char pw[255];
static int iport=6600;
static int waittime;

static mpd_Connection * conn;

static int errorcnt = 0;

#define MAX_CONV_LENGTH 256
char convStr[MAX_CONV_LENGTH];

/* -1 when mpd is not configured, 1 if mpd is initialized */
static int isconfigured=-1;

/* debug stuff, ripped from lcd4linux */
void message(const int level, const char *format, ...) __attribute__ ((format(__printf__, 2, 3)));
#define debug(args...) message (2, __FILE__ ": " args)
#define info(args...)  message (1, args)
#define error(args...) message (0, args) 

/* enable/disable debug messages here */
//#define DEBUGMSG

void message(const int level, const char *format, ...)
{

    va_list ap;
    char buffer[256];
    static int log_open = 0;
    
    va_start(ap, format);
    vsnprintf(buffer, sizeof(buffer), format, ap);
    va_end(ap); 

    switch (level) {
	case 0:
	    syslog(LOG_ERR, "%s", buffer);
	    break;
	case 1:
	    syslog(LOG_INFO, "%s", buffer);
	    break;
	default:
#ifdef DEBUGMSG
	    syslog(LOG_DEBUG, "%s", buffer);
#endif
	    break;
    } 

}

/* mpd stores all strings internal as utf8 string */
/* utf8 will display the first 128 characters like latin1 (iso-8859-1) characters */
/* but special chars will not be displayed correct (for example german umlauts) */
char * UTF8toLatin1 (char * from)
{
    char *p1, *p2;
    iconv_t i;
    size_t fromlen, tolen;
    size_t iconv_value;

    if (strlen(from) > MAX_CONV_LENGTH) {
	error("[bemusedMPD] ICONV ERROR1: max string size is %d, inputstring size: %d\n", MAX_CONV_LENGTH, strlen(from));
        return "ICONV ERROR1";
    }

    memset(convStr,0,sizeof(convStr));
    tolen = fromlen = strlen(from);
    p1 = (char *)from;
    p2 = (char *)convStr;

    i = iconv_open("ISO-8859-1", "UTF-8");
    if (i == (iconv_t) -1) {
        if (errno == EINVAL)
          error("[bemusedMPD] ICONV ERROR2: conversion from UTF-8 to ISO-8859-1 not available");
        else
	  error("[bemusedMPD] ICONV ERROR2: Conversion setup failed");
        return "ICONV ERROR2";
    }

    /* OpenWRT needs a type cast */
    /* iconv_value = iconv(i, (const char **)&p1, &fromlen, &p2, &tolen); */
    iconv_value = iconv(i, &p1, &fromlen, &p2, &tolen);
    if (iconv_value == (size_t) -1) {
        error ("ICONV ERROR3: failed to convert string, error: %d\n", errno);
	return "ICONV ERROR3";
    }

    iconv_close(i);
    return(convStr);
}


void close_mpd_connection() {
    mpd_closeConnection(conn);
}


int configure_mpd(void)
{
    char *test;
    
    if (isconfigured == 1)
	return 1;
	
    strcpy(host,"localhost");
    
    if (test = getenv("MPD_HOST"))
	strcpy(host, test);
    
    if (test = getenv("MPD_PORT"))
	iport = strtol(test, NULL, 10);	
    
    if ((iport < 0) || (iport > 65535)) {
	error("[bemusedMPD] MPD_PORT \"%s\" is not a positive integer - using default (6600)\n", iport);
	iport = 6600;
    }

    isconfigured = 1;
    return isconfigured; 
}

//check if connected
static int check_mpd_conn() {
    configure_mpd();
    
    if(conn==NULL || conn->error) {
	if (conn) {
	    if (errorcnt < ERROR_DISPLAY)
    		error("[Bemused-MPD] not connected, try to reconnect to [%s]:[%i], Error: [%s]\n", host, iport, conn->errorStr);
	    mpd_closeConnection(conn);
	}
	else
	    info("[Bemused-MPD] initialize connect to [%s]:[%i]\n", host, iport);

        conn = mpd_newConnection(host, iport, TIMEOUTINS);

        if(conn->error) {
	    if (errorcnt < ERROR_DISPLAY)
    		error("[Bemused-MPD] connection failed, give up. Reason: [%s]\n", conn->errorStr);
	    if (errorcnt == ERROR_DISPLAY)
		error("[Bemused-MPD] stop logging, until connection is fixed!");
	    errorcnt++;
    	    return -1;
        }
	
	if (errorcnt > ERROR_DISPLAY) {
	    error("[Bemused-MPD] connection is fixed!");
	}
	
	errorcnt = 0;
	info("[Bemused-MPD] connection fixed...\n");
    }

    return 0;
}

static int mpd_update()
{
    int ret = -1;    

    //check if connected
    if (check_mpd_conn()!=0)
	return -1;

#ifdef DEBUGMSG
    debug("[Bemused-MPD] mpd_update(): connection ok...");
#endif

    mpd_Status * status=NULL;
    mpd_Stats *stats=NULL;
    mpd_InfoEntity * entity;
    
    //generate query
    mpd_sendCommandListOkBegin(conn);
    mpd_sendStatsCommand(conn);

    if(conn->error) {
	error("[Bemused-MPD] error sendStat: %s\n",conn->errorStr);
	
	//try to fix connection. If a connection error occur, this will be the place!
	if (check_mpd_conn()!=0)
	    return -1;
	
	//retry connection
	mpd_sendCommandListOkBegin(conn);
	mpd_sendStatsCommand(conn);
	if(conn->error) {
	    error("[Bemused-MPD] error retry sendStat: %s\n",conn->errorStr);
	    return -1;
	}
    }

    mpd_sendStatusCommand(conn);
    mpd_sendCurrentSongCommand(conn);
    mpd_sendCommandListEnd(conn);
 
    //get stats
    stats = mpd_getStats(conn);
    if(stats==NULL) {
	error("[Bemused-MPD] error mpd_getStats: %s", conn->errorStr);
	goto cleanup;
    } 

    //get status
    mpd_nextListOkCommand(conn);
    if((status = mpd_getStatus(conn))==NULL) {
	error("[Bemused-MPD] error mpd_nextListOkCommand: %s", conn->errorStr);
	goto cleanup;
    }

    //get song information
    mpd_nextListOkCommand(conn);
    while((entity = mpd_getNextInfoEntity(conn))) {
	mpd_Song * song = entity->info.song;

	if (currentSong!=NULL)
    	    mpd_freeSong(currentSong);

	currentSong = mpd_songDup(song);
	mpd_freeInfoEntity(entity);
    }  
 
    l_elapsedTimeSec 	= status->elapsedTime;
    l_bitRate 		= status->bitRate;
    l_totalTimeSec 	= status->totalTime;
    l_repeatEnabled	= status->repeat;
    l_randomEnabled 	= status->random;
    l_state 		= status->state;
    l_volume 		= status->volume;
    l_playlistLength 	= status->playlistLength;
    l_currentSongPos 	= status->song;
    l_sampleRate 	= status->sampleRate;
    l_channels 		= status->channels;

    l_numberOfSongs 	= stats->numberOfSongs;
    l_uptime 		= stats->uptime;
    l_playTime 		= stats->playTime;
    l_dbPlayTime 	= stats->dbPlayTime;

    /* sanity checks */
    if (l_volume < 0 || l_volume > 100)
      l_volume = 0;

    if (l_bitRate < 0)
      l_bitRate = 0;
	
    ret = 0;
 
cleanup:
    if (stats!=NULL)
      mpd_freeStats(stats);
      
    if (status!=NULL)  
      mpd_freeStatus(status);

    if(conn->error) {
	error("[Bemused-MPD] error freeStatus: %s", conn->errorStr);
	return -1;
    }

    mpd_finishCommand(conn);
    if(conn->error) {
	error("[Bemused-MPD] error mpd_finishCommand: %s", conn->errorStr);
	return -1;
    }

    return ret; 
}


static mpd_Status * getStatus(mpd_Connection * conn) {
    mpd_Status * ret=NULL;

    mpd_sendStatusCommand(conn);
    if(conn->error) {
	error("[Bemused-MPD] error sendStatus: %s\n",conn->errorStr);
	return ret;
    }

    ret = mpd_getStatus(conn);
    if(conn->error) {
	error("[Bemused-MPD] error getStatus: %s\n",conn->errorStr);
	return ret;
    }

    mpd_finishCommand(conn);
    if(conn->error) {
	error("[Bemused-MPD] error finishCommand: %s\n",conn->errorStr);
    }

    return ret;
}
 
static void my_finishCommand(mpd_Connection * conn) {
    mpd_finishCommand(conn);
    if(conn->error) {
	error("[Bemused-MPD] error my_finishcommand: %s\n",conn->errorStr);
    }

} 

void mpd_remote_playlist_add(char *buffer, int isStream, char *basedir)
{    
	int i, num;
	int process=0;

	if (check_mpd_conn()==0) {
#ifdef DEBUGMSG
	    debug("add: %s, basedir: %s",buffer, basedir);
#endif
	    char * filenamebegin;
	    char filename[1024];
	    
	    //check for stream http://url
	    if (isStream == 1) {
		memcpy(filename, buffer, 1024);
		process=1;
	    } 
	    else {
		//check for filename
		filenamebegin=strstr(buffer, basedir);
		if (filenamebegin) {
		    //remove path from string, mpd just needs the songname
		    int len = strlen(basedir)+1;
		    memset(filename, 0, 1024);
		    memcpy(filename, buffer+len, strlen(buffer)-len);
		    process=1;
		}
	    }

	    if (process == 1) {
#ifdef DEBUGMSG
		debug("adding %s to playlist", filename);
#endif

		/* send list of songs */
		mpd_sendCommandListBegin(conn);
		mpd_sendAddCommand(conn,filename ); 
		if(conn->error) {
		    error("[Bemused-MPD] error addCommand: %s\n",conn->errorStr);
		}

		mpd_sendCommandListEnd(conn);
		my_finishCommand(conn);
	    }

	}
}

void mpd_remote_play()
{
    mpd_update(); 
    if (l_state == MPD_STATUS_STATE_PAUSE) {
        //resume from pause
	mpd_sendPauseCommand(conn, 0);
    } else {
        //start playing
        mpd_sendPlayCommand(conn, l_currentSongPos);
    }
    my_finishCommand(conn);
}

void mpd_remote_pause()
{
//toggle pause
    if (check_mpd_conn()==0) {
	mpd_update();
	if (l_state == MPD_STATUS_STATE_PAUSE) {
	    mpd_sendPauseCommand(conn, 0);
	} else {
    	    mpd_sendPauseCommand(conn, 1);
	}
        my_finishCommand(conn);
    }
}

void mpd_remote_stop()
{
    if (check_mpd_conn()==0) {
        mpd_sendStopCommand(conn);
	my_finishCommand(conn);
    }
}

gboolean mpd_remote_is_playing()
{
    mpd_update();
    if (l_state == MPD_STATUS_STATE_PLAY) {
#ifdef DEBUGMSG
	error("mpd_remote_is_playing=true");
#endif	
    	return (gboolean)true;
    }
    else {
#ifdef DEBUGMSG
	error("mpd_remote_is_playing=false");
#endif
    	return (gboolean)false;
    }
}

gboolean mpd_remote_is_paused()
{
    mpd_update();
    if (l_state == MPD_STATUS_STATE_PAUSE)
        return (gboolean)true;
    else
        return (gboolean)false;
}

/* id of current song in playlist */
gint mpd_remote_get_playlist_pos()
{
    mpd_update();
    return (gint)l_currentSongPos;
}

/* play another song in the playlist */
void mpd_remote_set_playlist_pos(int pos)
{
    if (check_mpd_conn()==0) {
	mpd_sendPlayCommand(conn, pos);
	my_finishCommand(conn);
    }
}

/* # of tracks in the current playlist */
gint mpd_remote_get_playlist_length()
{
    mpd_update();
    return (gint)l_playlistLength;
}

/* clear playlist */
void mpd_remote_playlist_clear()
{
    if (check_mpd_conn()==0) {
        mpd_sendClearCommand(conn);
	my_finishCommand(conn);	
    }
}

/* seek current track */
void mpd_remote_jump_to_time(int pos)
{
    mpd_update(); 
    if (check_mpd_conn()==0) {
	if (l_state == MPD_STATUS_STATE_PAUSE || l_state == MPD_STATUS_STATE_PLAY) {
	    mpd_sendSeekCommand(conn, l_currentSongPos, pos);
	    my_finishCommand(conn);
	}
    }
}

int mpd_remote_get_main_volume()
{
    mpd_update();
    return l_volume;
}

void mpd_remote_set_main_volume(int v)
{
    if (check_mpd_conn()==0) {
        mpd_sendSetvolCommand(conn, v);
	my_finishCommand(conn);
    }
}

//return song information for track pos
char *mpd_remote_get_playlist_title(int pos)
{
    static char retstr[255];
    static char title[127];
    static char artist[127];
    int len;

#ifdef DEBUGMSG
    debug("mpd_remote_get_playlist_title, get pos %i", pos);
#endif	
    if (check_mpd_conn()!=0)
	return "connection error!";

    memset(retstr, 0, 255);
    memset(title, 0, 127);
    memset(artist, 0, 127);

    //get track nr pos
    mpd_InfoEntity * entity;
 
    mpd_sendCommandListOkBegin(conn);
    mpd_sendPlaylistInfoCommand(conn, pos);
    mpd_sendCommandListEnd(conn);
 
    while((entity = mpd_getNextInfoEntity(conn))) {
	mpd_Song * song = entity->info.song;

        //tag check
        if (song->artist != NULL) {
	    len = strlen(song->artist);
    	    if (len > 126)
		len=126;
            memcpy(artist, song->artist, len);
    	    
	    if (song->title != NULL) {
	        len = strlen(song->title);
		if (len > 126)
    	            len=126;
		memcpy(title, song->title, len);
    		sprintf(retstr, "%s-%s", artist, title);
	    } else sprintf(retstr, "%s", artist);

#ifdef DEBUGMSG
	    debug("track: %s", retstr);
#endif
	} else {
    	    //no artist tag available.. mp3 without tag OR stream	    
    	    if (song->file != NULL) {
		len = strlen(song->file);
    		if (len > 126)
    	    	    len=126;
		memcpy(artist, song->file, len);
	    
		if (song->title != NULL) {
    	    	    len = strlen(song->title);
    	    	    if (len > 126)
			len=126;
		    memcpy(title, song->title, len);
		    sprintf(retstr, "%s-%s", artist, title);
		} else sprintf(retstr, "%s", artist);

#ifdef DEBUGMSG
	    debug("track: %s", retstr);
#endif	
	    }
	}
	mpd_freeInfoEntity(entity);
    }

    my_finishCommand(conn);

    return UTF8toLatin1( (char *)retstr );
}

/* current song position */
int mpd_remote_get_output_time()
{
    mpd_update();
    return l_elapsedTimeSec;
}

/* song length - pos is unused! */
int mpd_remote_get_playlist_time(int pos)
{
    mpd_update();
    return l_totalTimeSec;
}

void mpd_remote_get_info(int * rate, int * freq, int * nch)
{
	mpd_update();
	*rate = l_bitRate;
	*freq = l_sampleRate;
	*nch  = l_channels;
}

/* play prev song */
void mpd_remote_playlist_prev()
{
    if (check_mpd_conn()==0) {
	mpd_sendPrevCommand(conn); 
	my_finishCommand(conn);	
    }
}

/* play next song */
void mpd_remote_playlist_next()
{
    if (check_mpd_conn()==0) {
        mpd_sendNextCommand(conn);
	my_finishCommand(conn);
    }
}

/* check if audio player is running */
gboolean mpd_remote_is_running()
{
    if (check_mpd_conn() == 0) 
	return true;

    return false;
}


void mpd_remote_toggle_shuffle()
{
    int mode;

    if (check_mpd_conn()==0) {	    	
        mpd_Status * status;
	status = getStatus(conn);
        mode = !status->random;
	mpd_freeStatus(status);

        mpd_sendRandomCommand(conn,mode);
	my_finishCommand(conn);
    }
}


gboolean mpd_remote_is_repeat()
{
    mpd_update(); 
    if (l_repeatEnabled==1)
        return (gboolean)true;
    else 
        return (gboolean)false;

}

gboolean mpd_remote_is_shuffle()
{
    mpd_update(); 
    if (l_randomEnabled==1)
        return (gboolean)true;
    else 
	return (gboolean)false;
}

void mpd_remote_toggle_repeat()
{
    int mode;

    if (check_mpd_conn()==0) {
        mpd_Status * status;
        status = getStatus(conn);
        mode = !status->repeat;
	mpd_freeStatus(status);

        mpd_sendRepeatCommand(conn,mode);
	my_finishCommand(conn);
    }
}
